using System;
using System.Diagnostics;
using System.Globalization;
using System.ComponentModel;
using System.Configuration;
using System.Windows;
using System.Windows.Markup;
using System.Windows.Forms;
using System.Drawing;

namespace Yorrick.Common
{
    /// <summary>
    /// Persists a Window's Size, Location and WindowState to UserScopeSettings 
    /// </summary>
    internal class WindowSettings
    {
        #region WindowApplicationSettings Helper Class
        public class WindowApplicationSettings : ApplicationSettingsBase
        {
            private WindowSettings windowSettings;

            public WindowApplicationSettings(WindowSettings windowSettings)
                : base(windowSettings.window.Name)
            {
                this.windowSettings = windowSettings;
            }

            [UserScopedSetting]
            public Rect Location
            {
                get
                {
                    if (this["Location"] != null)
                    {
                        return ((Rect)this["Location"]);
                    }
                    return Rect.Empty;
                }
                set
                {
                    this["Location"] = value;
                }
            }

            [UserScopedSetting]
            public WindowState WindowState
            {
                get
                {
                    if (this["WindowState"] != null)
                    {
                        return (WindowState)this["WindowState"];
                    }
                    return WindowState.Normal;
                }
                set
                {
                    this["WindowState"] = value;
                }
            }

        }
        #endregion

        #region Constructor
        private Window window = null;

        public WindowSettings(Window window)
        {
            this.window = window;
        }

        #endregion

        #region Attached "Save" Property Implementation
        /// <summary>
        /// Register the "Save" attached property and the "OnSaveInvalidated" callback 
        /// </summary>
        public static readonly DependencyProperty SaveProperty
           = DependencyProperty.RegisterAttached("Save", typeof(bool), typeof(WindowSettings),
                new FrameworkPropertyMetadata(new PropertyChangedCallback(OnSaveInvalidated)));

        public static void SetSave(DependencyObject dependencyObject, bool enabled)
        {
            dependencyObject.SetValue(SaveProperty, enabled);
        }

        /// <summary>
        /// Called when Save is changed on an object.
        /// </summary>
        private static void OnSaveInvalidated(DependencyObject dependencyObject, DependencyPropertyChangedEventArgs e)
        {
            Window window = dependencyObject as Window;
            if (window != null)
            {
                if ((bool)e.NewValue)
                {
                    WindowSettings settings = new WindowSettings(window);
                    settings.Attach();
                }
            }
        }

        #endregion

        #region Protected Methods
        /// <summary>
        /// Load the Window Size Location and State from the settings object
        /// </summary>
        protected virtual void LoadWindowState()
        {
            this.Settings.Reload();
            if (this.Settings.Location != Rect.Empty)
            {
                TryRestoreWindowLocation(this.Settings.Location);
            }

            if (this.Settings.WindowState != WindowState.Maximized)
            {
                this.window.WindowState = this.Settings.WindowState;
            }
        }

        /// <summary>
        /// Tries to restore the windows position to the specified location.
        /// </summary>
        /// <param name="windowLocation">The window location to restore to.</param>
        /// <remarks>The window location is only used if it falls within
        /// the visible region of one of the attached screens.</remarks>
        private void TryRestoreWindowLocation(Rect windowLocation)
        {
            this.Settings.Reload();
            if (this.Settings.Location != Rect.Empty)
            {
                bool isInOneOfTheScreens = false;
                foreach (Screen s in Screen.AllScreens)
                {
                    isInOneOfTheScreens |= IsWindowRectInScreenRect(windowLocation, ToRect(s.WorkingArea));
                }

                if (isInOneOfTheScreens)
                {
                    this.window.Left = windowLocation.Left;
                    this.window.Top = windowLocation.Top;
                    this.window.Width = windowLocation.Width;
                    this.window.Height = windowLocation.Height;
                }
            }
        }

        /// <summary>
        /// Converts a <see cref="Rectangle"/> to a <see cref="Rect"/> and returns the result.
        /// </summary>
        /// <param name="rectangle">The <see cref="Rectangle"/> to convert.</param>
        /// <returns>The resulting <see cref="Rect"/> from the conversion.</returns>
        private static Rect ToRect(Rectangle rectangle)
        {
            return new Rect
            {
                X = rectangle.Left,
                Y = rectangle.Top,
                Height = rectangle.Height,
                Width = rectangle.Width
            };
        }  

        /// <summary>
        /// Determines if the window lies within a screen.
        /// </summary>
        /// <param name="windowRect">The <see cref="Rect"/> for the window.</param>
        /// <param name="screenRect">The <see cref="Rect"/> for the screen.</param>
        private static bool IsWindowRectInScreenRect(Rect windowRect, Rect screenRect)
        {
            return screenRect.Contains(windowRect);
        }

        /// <summary>
        /// Save the Window Size, Location and State to the settings object
        /// </summary>
        protected virtual void SaveWindowState()
        {
            this.Settings.WindowState = this.window.WindowState;
            this.Settings.Location = this.window.RestoreBounds;
            this.Settings.Save();
        }
        #endregion

        #region Private Methods

        private void Attach()
        {
            if (this.window != null)
            {
                this.window.Closing += new CancelEventHandler(window_Closing);
                this.window.Initialized += new EventHandler(window_Initialized);
                this.window.Loaded += new RoutedEventHandler(window_Loaded);
            }
        }

        private void window_Loaded(object sender, RoutedEventArgs e)
        {
            if (this.Settings.WindowState == WindowState.Maximized)
            {
                this.window.WindowState = this.Settings.WindowState;
            }
        }

        private void window_Initialized(object sender, EventArgs e)
        {
            LoadWindowState();
        }

        private void window_Closing(object sender, CancelEventArgs e)
        {            
            SaveWindowState();
        }
        #endregion

        #region Settings Property Implementation
        private WindowApplicationSettings windowApplicationSettings;

        protected virtual WindowApplicationSettings CreateWindowApplicationSettingsInstance()
        {
            return new WindowApplicationSettings(this);
        }

        [Browsable(false)]
        public WindowApplicationSettings Settings
        {
            get
            {
                if (windowApplicationSettings == null)
                {
                    this.windowApplicationSettings = CreateWindowApplicationSettingsInstance();
                }
                return this.windowApplicationSettings;
            }
        }
        #endregion
    }
}

